/*-
 * Copyright (C)2013 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2013 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: anypost.c,v 1.1 2013/03/23 11:20:58 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sysexits.h>
#include <sys/uio.h>
#include <strings.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <limits.h>
#include <netdb.h>
#include <stdio.h>
#include <errno.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <mife.h>

#define CRV 256
#define WRV 7
#define HTVER 2

typedef enum
{   anypost_none = 0
,   anypost_http = 1
} anypost_sh;

typedef struct {
    BLIN_flag  flags;
#       define ANYPOST_SEEPOST 0x800000
#       define ANYPOST_CONTLEN 0x400000
    anypost_sh sh;
    char      *host;
    char      *port;
    char      *user;
    char      *pass;
    char      *get;
    int        proto;
} anypost_cf;

typedef struct {
    char            *f;
    char            *a;
} anypost_header;

static char *empty = "";
static char *get = "POST ";
static char *end = "\r\n";
static char *httpver[] =
{ ""
, " HTTP/1.0\r\nHost: "
, " HTTP/1.1\r\nHost: "
};

static void
usage(int ex) {
    fprintf( stderr
           , "anypost " VERS "  " DATE "\n"
             "    -c        - no Content-Length: header\n"
             "    -d        - duplicate request to stderr\n"
             "    -F file   - add headers from file\n"
             "    -h header - add header\n"
             "    -h        - HTTP version\n"
             "    -q        - quiet\n"
             "    -v        - verbose\n"
           )
    ;
    exit(ex);
}

static char *
rediropt(anypost_cf *options, babolo_opt **opt, int redirenv) {
    ifBLIN_FLAG = options->flags;
    char *e, *o;

    if  (!(o = babolo_getoptarg(opt, 0))) {
        ifBLIN_QW0("No argument");
        usage(EX_USAGE);
    } else if ((redirenv > 0) && (!(o = getenv(e = o)))) {
        ifBLIN_QX0("No ENV %s", e);
        usage(EX_USAGE);
    }
    return(o);
}

int
main(int argc, char *argv[], char **envp) {
#   define blin_internal_flags (options->flags)
    int c, s, w, redirenv, error, crv, wrv, ex = EX_OK;
    struct addrinfo *hint, *res, *res0;
    mife_descriptor *mdi, *mdh;
    size_t respl, wrto, wrin;
    anypost_cf *options;
    char *flar, *o, *p;
    anypost_header *cr;
    BLIN_flag subflags;
    struct iovec *buf;
    ssize_t postsize;
    babolo_parm *bp;
    babolo_opt *opt;
    void *resp;

    if  (!(options = calloc(1, sizeof(*options)))) {
        warnx("Malloc failed #1");
        errno = ENOMEM;
        ex = EX_OSERR;
        goto out;
    }
    options->proto = HTVER;
    options->flags = BLIN_VER1;
    if  (!(cr = calloc(CRV, sizeof(*cr)))) {
        ifBLIN_QX0("Malloc failed");
        errno = ENOMEM;
        ex = EX_OSERR;
        goto out;
    }
    if  (babolo_openopt(Bpars_NONU, &opt, argc, argv, flar = "cdEF:f:h:qvz:Z:?")) {
        ifBLIN_QW0("openopt");
        ex = EX_SOFTWARE;
        goto out;
    }
    crv = 0;
    for (redirenv = 0; (c = babolo_getopt(&opt)) != 0; redirenv--) {
        if  (redirenv < 0) redirenv = 0;
        switch (c) {
        case 'c': options->flags |= ANYPOST_CONTLEN;                                             break;
        case 'd': options->flags |= ANYPOST_SEEPOST;                                             break;
        case 'E': redirenv = 2;                                                                  break;
        case 'F': if  (crv >= CRV) {
                ;     ifBLIN_QX0("Too many headers, %d max", CRV);
                ;     usage(EX_USAGE);
                ; }
                ; cr[crv++].f = rediropt(options, &opt, redirenv);
                ; break;
        case 'f': if  (crv >= CRV) {
                ;     ifBLIN_QX0("Too many headers, %d max", CRV);
                ;     usage(EX_USAGE);
                ; }
                ; cr[crv++].a = rediropt(options, &opt, redirenv);
                ; break;
        case 'h': o = rediropt(options, &opt, redirenv);
                ; options->proto = strtoul(o, NULL, 0);
                ; if  (options->proto > HTVER) {
                ;     ifBLIN_QX1("htver %d > %d", options->proto, HTVER);
                ;     options->proto = HTVER;
                ; }
                ; break;
        case 'q': options->flags &= ~BLIN_VERMASK;                                               break;
        case 'v': BLIN_VERBOSE(options->flags);                                                  break;
        case 'z': o = rediropt(options, &opt, redirenv);
                ; bp = babolo_getparm(Bpars_NOEN | Bpars_NOAL | Bpars_NONU, (u_char**)&o, (u_char*)"", 0);
                ; babolo_openopt(0, &opt, bp->argc, (char **)bp->argv, NULL);
                ; break;
        case 'Z': o = rediropt(options, &opt, redirenv);
                ; bp = babolo_getparm(Bpars_NOAL | Bpars_NONU, (u_char**)&o, (u_char*)"", 0);
                ; babolo_openopt(0, &opt, bp->argc, (char **)bp->argv, NULL);
                ; break;
        case '?': usage(EX_OK);
        default : ifBLIN_QX0("Illegal flag -%c", c);
                ; usage(EX_USAGE);
    }   }
    if  (!(o = babolo_getarg(&opt))) {
        ifBLIN_QW0("URL absent");
        ex = EX_USAGE;
        goto out;
    }
    subflags = (((BLIN_VER1 | BLIN_VER2 ) & options->flags) | (options->flags >> 1)) & BLIN_VERMASK;
    if  (strncmp(o, "http://", strlen("http://"))) {
        ifBLIN_QX0("URL=%s~: Illegal shema", o);
        errno = EINVAL;
        ex = EX_USAGE;
        goto out;
    }
    options->sh = anypost_http;
    o += strlen("http");
    *o++ = '\0';
    o += strlen("//");
    if  ((p = strpbrk(o, "/@")) && *p == '@') {
        /* username */
        *p++ = '\0';
        options->user = o;
        if  ((o = index(o, ':'))) {
            *o++ = '\0';
            options->pass = o;
        }
        o = p;
    }
    options->host = o;
    if  ((p = strpbrk(o, "/:")) && *p == ':') {
        /* port */
        *p++ = '\0';
        options->port = p;
        o = p;
    }
    if  (o && (o = index(o, '/'))) *o++ = '\0';
    options->get = o ? o : empty;
    if  (!options->port) options->port = "80";
    ifBLIN_QV3(options->flags) {
        ifBLIN_QX2("sh=%d proto=%d\n", options->sh, options->proto);
        if  (options->user) ifBLIN_QX2("user=%s~\n", options->user);
        if  (options->pass) ifBLIN_QX2("pass=%s~\n", options->pass);
        if  (options->host) ifBLIN_QX2("host=%s~\n", options->host);
        if  (options->port) ifBLIN_QX2("port=%s~\n", options->port);
        if  (options->get) ifBLIN_QX2("get=%s~\n", options->get);
        for (c = 0; c < crv; ++c) {
            if  (cr[c].f) ifBLIN_QX2("f[%d]=%s~\n", c, cr[c].f);
            if  (cr[c].a) ifBLIN_QX2("a[%d]=%s~\n", c, cr[c].a);
    }   }
    if  (!(hint = calloc(1, sizeof(*hint)))) {
        ifBLIN_QX0("Malloc failed");
        errno = ENOMEM;
        ex = EX_OSERR;
        goto out;
    }
    hint->ai_family = PF_INET;
    hint->ai_socktype = SOCK_STREAM;
    hint->ai_protocol = IPPROTO_TCP;
    hint->ai_flags = AI_ADDRCONFIG;
    if  ((error = getaddrinfo(options->host, options->port, hint, &res))) {
        ifBLIN_QX0("%d=%s", error, gai_strerror(error));
        errno = ENETUNREACH;
        ex = EX_NOHOST;
        goto out;
    }
    for (s = -1, res0 = res; res; res = res->ai_next) {
        if  (0 > (s = socket(res->ai_family, res->ai_socktype, res->ai_protocol))) {
            ifBLIN_QW1( "socket=%d family=%d socktype=%d protocol=%d fail"
                      , s, res->ai_family, res->ai_socktype, res->ai_protocol
                      );
            continue;
        }
        if  (0 > (connect(s, res->ai_addr, res->ai_addrlen))) {
            ifBLIN_QW1("connect s=%d", s);
            close(s);
            s = -1;
            continue;
        }
        break;
    }
    if  (s < 0) {
        ifBLIN_QX0("Not connected %s:%s", options->host, options->port);
        ex = EX_UNAVAILABLE;
        goto out;
    }
    freeaddrinfo(res0);
    for (wrv = WRV, c = 0; c < crv; ++c) {
        if  (cr[c].f) ++wrv;
        if  (cr[c].a) wrv += 2;
    }
    if  (!(buf = calloc(wrv, sizeof(*buf)))) {
        ifBLIN_QX0("Malloc failed");
        errno = ENOMEM;
        ex = EX_OSERR;
        goto out;
    }
    w = 0;
    buf[w].iov_base = get;
    buf[w++].iov_len = strlen(get);
    buf[w].iov_base = "/";
    buf[w++].iov_len = 1;
    buf[w].iov_base = options->get;
    buf[w++].iov_len = strlen(options->get);
    if  (options->proto) {
        buf[w].iov_base = httpver[options->proto];
        buf[w++].iov_len = strlen(httpver[options->proto]);
        buf[w].iov_base = options->host;
        buf[w++].iov_len = strlen(options->host);
        buf[w].iov_base = end;
        buf[w++].iov_len = strlen(end);
    }
    for (c = 0; c < crv; ++c) {
        if  (cr[c].f) {
            if  (!(mdi = mife_open(subflags, cr[c].f))) {
                ifBLIN_QW0("mife_open -F=%s~", cr[c].f);
                ex = EX_IOERR;
                goto out;
            }
            if  (0 > (buf[w].iov_len = mife_read(mdi, 0, 0))) {
                ifBLIN_QW0("mife_read -F=%s~");
                ex = EX_IOERR;
                goto out;
            }
            if  (!(o = mife_get(mdi, 0))) {
                ifBLIN_QW0("mife_get -F=%s~");
                ex = EX_IOERR;
                goto out;
            }
            buf[w].iov_base = strndup(o, buf[w].iov_len);
            mife_close(mdi);
            mdi = NULL;
            ++w;
        }
        if  (cr[c].a) {
            buf[w].iov_base = cr[c].a;
            buf[w++].iov_len = strlen(cr[c].a);
            buf[w].iov_base = end;
            buf[w++].iov_len = strlen(end);
    }   }
    if  (!(mdi = mife_opef(subflags, fileno(stdin)))) {
        ifBLIN_QW0("mife_opef stdin");
        ex = EX_IOERR;
        goto out;
    }
    if  (0 > (postsize = mife_read(mdi, 0, 0))) {
        ifBLIN_QW0("mife_read stdin");
        ex = EX_IOERR;
        goto out;
    }
    if  (!(options->flags & ANYPOST_CONTLEN)) {
        if  (0 > (buf[w].iov_len = asprintf((void*)&buf[w].iov_base, "Content-Length: %u", postsize))) {
            ifBLIN_QW0("asprintf failed");
            errno = ENOMEM;
            ex = EX_OSERR;
            goto out;
        }
        buf[++w].iov_base = end;
        buf[w++].iov_len = strlen(end);
    }
    buf[w].iov_base = end;
    buf[w++].iov_len = strlen(end);
    buf[w].iov_len = postsize;
    if  (!(buf[w++].iov_base = mife_get(mdi, 0))) {
        ifBLIN_QW0("mife_get stdin");
        ex = EX_IOERR;
        goto out;
    }
    for (wrto = 0, c = 0; c < w; ++c) {
        ifBLIN_QX2("[%d]=%s~\n", c, buf[c].iov_base);
        wrto += buf[c].iov_len;
    }
    if  ((options->flags & ANYPOST_SEEPOST) && (wrto != (wrin = writev(fileno(stderr), buf, w)))) {
        ifBLIN_QW0("writev debug %d <> %d", wrin, wrto);
        ex = EX_IOERR;
        goto out;
    }
    if  (wrto != (wrin = writev(s, buf, w))) {
        ifBLIN_QW0("writev %d <> %d", wrin, wrto);
        ex = EX_IOERR;
        goto out;
    }
    mife_close(mdi);
    mdi = NULL;
    if  (!(mdh = mife_opef(subflags, s))) {
        ifBLIN_QW0("mife_opef socket");
        ex = EX_IOERR;
        goto out;
    }
    if  (0 > (respl = mife_read(mdh, 0, 0))) {
        ifBLIN_QW0("mife_read socket");
        ex = EX_IOERR;
        goto out;
    }
    if  (!(resp = mife_get(mdh, 0))) {
        ifBLIN_QW0("mife_get socket");
        ex = EX_IOERR;
        goto out;
    }
    if  (respl != mife_writ(fileno(stdout), resp, respl)) {
        ifBLIN_QW0("mife_writ");
        ex = EX_IOERR;
        goto out;
    }
;
out:
    exit(ex);
}
